#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <string.h>

#include <gtk/gtkgc.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>


#include "network.h"
#include "waterfall.h"

static int make_socket (unsigned short int port);
static network_client *network_client_from_fd(int fd);
static void network_client_close (network_client *nc);
static network_client *network_client_new(void);
static void network_client_free(network_client *);
static int network_client_flush_fifo(network_client *nc);
static void network_client_output_fifo_add (network_client *nc, char *buffer, int len, int maxBytesQueued);
static int network_client_write1 (network_client *nc, char *buffer, int len, network_blocking_behavior ifBlock);
static network_client_output_fifo *network_client_output_fifo_new(char *newbuf, int len);
static void network_client_output_fifo_free(network_client_output_fifo *nf);
static void network_close_or_flush();
static int network_handle_one(network_client *nc, char c);
extern void network_init_client(network_client *nc);
void network_abort(void);

static fd_set network_active_fd_set, read_network_fd_set;
static int network_server_socket;

// We need a path to communicate current client to rest of code.
network_client *network_clients;

static gchar *network_tx_authorization_password = NULL;

void network_init(gchar *password) {
  network_tx_authorization_password = password;
  /* Create the socket and set it up to accept connections. */
  network_server_socket = make_socket (PORT);
  if (listen (network_server_socket, 1) < 0) {
    perror ("listen");
    exit (EXIT_FAILURE);
  }
  fprintf(stderr, "network: listening on port %d\n", PORT);

  /* Initialize the set of active sockets. */
  FD_ZERO (&network_active_fd_set);
  FD_SET (network_server_socket, &network_active_fd_set);
}


int network_poll(char *buffer, int len, network_client **ncptr) {
  struct timeval timeoutImmediately;
  timeoutImmediately.tv_sec = 0;
  timeoutImmediately.tv_usec = 0;

  *ncptr = NULL;

  /* Poll for input to arrive on one or more active sockets. */
  read_network_fd_set = network_active_fd_set;
  if (select (FD_SETSIZE, &read_network_fd_set, NULL, NULL, &timeoutImmediately) < 0) {
    perror ("select");
    exit (EXIT_FAILURE);
  }

  int fd;
  
  /* Service all the sockets with input pending. */
  for (fd = 0; fd < FD_SETSIZE; fd++) {
    if (FD_ISSET (fd, &read_network_fd_set)) {
      if (fd == network_server_socket) {
	/* Connection request on server listenting socket. */
	network_client *nc = network_client_new();
	size_t size = sizeof (*(nc->client_sockaddr_in));
	int clientSocket = accept (network_server_socket,
				   (struct sockaddr *) nc->client_sockaddr_in,
				   &size);
	if (clientSocket < 0) {
	  perror ("accept");
	  // got error -- return -1;
	  return -1;
	}
	// O_NONBLOCK is posix, and recv with MSG_PEEK returns -1 and errno=EAGAIN in case of no data
//	if (fcntl(clientSocket, F_SETFL, O_NONBLOCK) < 0) {
//	  perror("fcntl O_NON_BLOCK");
//	  return -1;
//	}
	// O_NDELAY is not posix, and recv with MSG_PEEK returns 0 in case of no data
	if (fcntl(clientSocket, F_SETFL, O_NDELAY) < 0) {
	  perror("fcntl O_NDELAY");
	  return -1;
	}
	fprintf (stderr,
		 "network: connect from host %s, port %hd socket %d\n",
		 inet_ntoa (nc->client_sockaddr_in->sin_addr),
		 ntohs (nc->client_sockaddr_in->sin_port),
		 clientSocket);
	FD_SET (clientSocket, &network_active_fd_set);
	nc->fd = clientSocket;
	nc->next = network_clients;
	network_clients = nc;
	*ncptr = nc;	/* evil we accept only one new client per poll */
	nc->state = OPEN;
	network_init_client(nc);
	if (0) fprintf(stderr, "new client seen\n");
	return 0;
      } else {
	/* Data arriving on a client socket. */
	if (0) fprintf(stderr, "Data arriving on a client socket\n");
	int nbytes;
	network_client *nc = network_client_from_fd(fd);
	nbytes = recv (fd, buffer, len, MSG_PEEK);
	if (0) {
	  if (nbytes >0) 
	    fprintf(stderr, "MSG_PEEK returned %d bytes on %d errno = %d data %d\n", nbytes, fd, errno, buffer[0]);
	  else
	    fprintf(stderr, "MSG_PEEK returned %d bytes on %d errno = %d\nn", nbytes, fd, errno);
	}
	if (nbytes < 0) {
	  fprintf(stderr, "read %d bytes so closing\n", nbytes);
	  // got no data but lost client.
	  perror("recv");
	  nc->state = CLOSING;
	  *ncptr = NULL;
	  return 0;
	} else {
	  if (0) fprintf(stderr, "read %d bytes \n", nbytes);
	  // got data -- return number of bytes.
	  // Data is left in queue.
	  *ncptr = nc;
	  return nbytes;
	}
      }
    }
  }
  // no data, no error, return 0;
  return 0;
}
void network_update_mode(char *mode_name) {
  if (network_clients != NULL) {
    if (0) fprintf(stderr, "Network update mode %s\n", mode_name);
    int nbytes = 0;
    char cmd = 'M';
    char buffer[257];
    int len = strlen(mode_name);
    if (len > 255) {
      fprintf(stderr, "Mode name length > 255: %d %s\n", len, mode_name);
      return;
    }
    network_client *nc;
    buffer[nbytes++] = cmd;
    buffer[nbytes++] = len;
    memcpy(buffer+nbytes, mode_name, len);
    nbytes += len;
    for (nc = network_clients; nc != NULL; nc = nc->next) {
      if (nc->last_mode == NULL || strcmp(mode_name, nc->last_mode) != 0) {
	nc->last_mode = mode_name; /* depend on string being permanently allocated */
	network_client_write(nc, buffer, nbytes, QUEUE_IF_BLOCK);
      }
    }
  }
}



void waterfall_external_set_frequency(GtkWidget *widget, int x);
extern int waterfall_last_frequency; /* crock evil hack */
// crock -- we send F to all clients when a new one joins.
extern GtkWidget *waterfall;

void network_process() {
  int done = 0;
  while (!done) {
    // close or flush any clients synchronously
    network_close_or_flush();

    // Poll for input or new clients
    {
      char c;
      network_client *nc = NULL;
      // network_poll leaves data in queue - we just get a peek.
      int nbytes = network_poll(&c, 1, &nc);
      if ((nc != NULL) && nc->state == NEW)
	network_init_client(nc);
      if (nbytes < 1) 
	break;
      done = network_handle_one(nc, c);
    }
  }
}

void network_close_or_flush() {
  network_client *nc = network_clients;
  // Flush out any open clients
  while (nc != NULL) {
    //fprintf(stderr, "Getting lock on nc->fd %d\n", nc->fd);
    //g_mutex_lock(nc->mutex);
    //fprintf(stderr, "Got lock on nc->fd %d\n", nc->fd);
    if (nc->state == OPEN) {
      // fprintf(stderr, "Was open nc->fd %d\n", nc->fd);
      if ((network_client_flush_fifo(nc)) < 0) {
	fprintf(stderr, "flush_fifo failed nc->fd %d\n", nc->fd);
	nc->state = CLOSING;
	if (nc->tx) {
	  fprintf(stderr, "tx was on nc->fd %d\n", nc->fd);
	  // This causes a deadlock with the waterfall mutex.  Why?
	  fprintf(stderr, "A TX client is closing; pushing the abort button.");
	  network_abort();
	} 
      } 
      //fprintf(stderr, "unlocking mutex on nc->fd %d\n", nc->fd);
      //g_mutex_unlock(nc->mutex);
      //fprintf(stderr, "unlocked mutex on nc->fd %d\n", nc->fd);
      nc = nc->next;
    }
  }
  // Close any broken clients
  nc = network_clients;
  while (nc != NULL) {
    network_client *next = nc->next;
    if (nc->state == CLOSING) {
      network_client_close(nc);
    }
    nc = next;
  }
}

int network_handle_one(network_client *nc, char c) {
  int done = 0;
  int nbytes = 0;
  fprintf(stderr, "Handling %d for %d\n", c, nc->fd);
  switch(c) {
  case 'A': 
    {
      unsigned char a[1];
      nbytes = network_read_from_client (nc, a, 1);
      if (nbytes < 0) {
	done = 1;
	break;
      }
      fprintf(stderr, "Ack\n");
      break;
    }

  case 'S': 
    {
      unsigned char s[3];
      nbytes = network_read_from_client (nc, s, 3);
      if (nbytes < 3) {
	done = 1;
	break;
      }
      int waterfall_width = ((unsigned char)s[1])*256+((unsigned char)s[2]);
      nc->waterfall_width = waterfall_width;
      fprintf(stderr, "Set waterfall width to %d\n", waterfall_width);
      break;
    }

  case 'M': 
    {
      extern void set_mode_name(char *);
      unsigned char l[2];
      unsigned char mode[256];
      nbytes = network_read_from_client (nc, l, 2);
      if (nbytes < 2) {
	done = 1;
	break;
      }
      int len = l[1];		/* l[0] == 'M' */
      nbytes = network_read_from_client (nc, mode, len);
      if (nbytes < len) {
	done = 1;
	break;
      }
      mode[len] = '\0';
      set_mode_name(mode);
      break;
    }

  case 'X': 
    {
      unsigned char l[2];
      unsigned char login[256];
      nbytes = network_read_from_client (nc, l, 2);
      if (nbytes < 2) {
	done = 1;
	break;
      }
      int len = l[1];		/* l[0] == 'X' */
      if (len > 0) {
	nbytes = network_read_from_client (nc, login, len);
	if (nbytes < len) {
	  done = 1;
	  break;
	}
	login[len] = '\0';
	fprintf(stderr, "Client %s trying to authorize for TX\n", inet_ntoa(nc->client_sockaddr_in->sin_addr));
	if (network_tx_authorization_password == NULL) {
	  fprintf(stderr, "No TX password is set!\n");
	}
	if (network_tx_authorization_password != NULL && strcmp(login, network_tx_authorization_password) == 0) {
	  fprintf(stderr, "Client %s authorized for TX\n", inet_ntoa(nc->client_sockaddr_in->sin_addr));
	  nc->tx_authorized = 1;
	  // probably should turn off TX bit on other clients now.
	} else {
	  fprintf(stderr, "Client %s failed to authorize for TX\n", inet_ntoa(nc->client_sockaddr_in->sin_addr));
	  nc->tx_authorized = 0;
	  // should probably make sure client is not in xmit now.
	}
      } else {
	fprintf(stderr, "Client %s closed authorization.\n", inet_ntoa(nc->client_sockaddr_in->sin_addr));
	nc->tx_authorized = 0;
	if (nc->tx) {
	  fprintf(stderr, "Client %s was in TX; aborting TX.\n", inet_ntoa(nc->client_sockaddr_in->sin_addr));
	  network_abort();
	}
      }
      break;
    }

  case 'R': 
    {
      extern void push_button(const gchar *name);
      unsigned char l[2];
      unsigned char button[256];
      nbytes = network_read_from_client (nc, l, 2);
      if (nbytes < 2) {
	done = 1;
	break;
      }
      int len = l[1];		/* l[0] == 'R' */
      nbytes = network_read_from_client (nc, button, len);
      if (nbytes < len) {
	done = 1;
	break;
      }
      button[len] = '\0';
      // CROCK: We know what buttons require TX authorization.
      if (!nc->tx_authorized &&
	  ((strcmp(button, "txbutton") != 0) || 
	   (strcmp(button, "rxbutton") != 0) || 
	   (strcmp(button, "abortbutton") != 0))) {
	fprintf(stderr, "Client %s not authorized to push button %s\n", inet_ntoa(nc->client_sockaddr_in->sin_addr), button);
      } else {
	fprintf(stderr, "Client %s pushing button %s\n", inet_ntoa(nc->client_sockaddr_in->sin_addr), button);
	if (strcmp(button, "txbutton") != 0) {
	  nc->tx = 1;
	}
	push_button(button);
	if (strcmp(button, "rxbutton") != 0 || strcmp(button, "abortbutton") != 0) {
	  nc->tx = 0;
	}
      }
      break;
    }

  case 'T': 
    {
      extern void network_insert_tx_text(gchar *, int len);
      unsigned char l[2];
      gchar text[256];
      nbytes = network_read_from_client (nc, l, 2);
      if (nbytes < 2) {
	done = 1;
	break;
      }
      int len = l[1];		/* l[0] == 'T' */
      fprintf(stderr, "T %d\n", len);
      nbytes = network_read_from_client (nc, text, len);
      fprintf(stderr, "read %d bytes\n", nbytes);
      if (nbytes < len) {
	done = 1;
	break;
      }
      text[len] = '\0';
      network_insert_tx_text(text, len);
      break;
    }

  case 'F': 
    {
      unsigned char f[3];
      nbytes = network_read_from_client (nc, f, 3);
      if (nbytes < 3) {
	fprintf(stderr, "could not read freq \n");
	done = 1;
	break;
      }
      {
	int x2 = ((unsigned char)f[1])*256+((unsigned char)f[2]);
	if (0) fprintf(stderr, "Setting waterfall freq to x2=%d\n", x2);
	waterfall_external_set_frequency(((GtkWidget *)waterfall), x2);
      }
      break;
    }

  case 'P': 
    {
      unsigned char s[3];
      nbytes = network_read_from_client (nc, s, 2);
      if (nbytes < 2) {
	done = 1;
	break;
      }
      int miniscope_skip_amount = ((unsigned char)s[1]);
      nc->miniscope_skip_amount = miniscope_skip_amount;
      break;
    }
    
  case '\0':
    // When this happens, it means that MSG_PEEK has failed on us and returned 1 for len but no data.
    // This never gets any better.
    fprintf(stderr, "MSG_PEEK has failed on us and returned 1 for len but data buffer of 0.\n");
    {
      unsigned char s[4];
      nbytes = network_read_from_client (nc, s, 1);
      if (nbytes == 1) 
	fprintf(stderr, "read byte %d\n", s[0]);
      else fprintf(stderr, "could not read another byte\n");
    }
    // nc->state = CLOSING;
    // done = TRUE;
    break;

  default:
    fprintf(stderr, "unknown char %c (%d)\n", c, c);
    done = TRUE;
  }
  return done;
}

void network_abort() {
  extern void push_button(const gchar *name);
  push_button("abortbutton");
  fprintf(stderr, "Pushed the abort button.");
}

int network_client_write (network_client *nc, char *buffer, int len, network_blocking_behavior ifBlock) {
  int nbytes;
  g_mutex_lock(nc->mutex);
  nbytes = network_client_write1(nc, buffer, len, ifBlock);
  g_mutex_unlock(nc->mutex);
  return nbytes;
}

static int network_client_write1 (network_client *nc, char *buffer, int len, network_blocking_behavior ifBlock) {
  if (nc->state == CLOSING) {
    fprintf(stderr, "Ignoring write on %d because client closing\n", nc->fd);
    return len;
  }

  int fd = nc->fd;

  int nbytes = 0;
  switch(ifBlock) {
  case BLOCK_IF_BLOCK:
    if (0) fprintf(stderr, "Blocking on writing %d %c...", len, buffer[0]);
    fflush(stderr);
    nbytes = send(fd, buffer, len, 0);
    nc->network_writes_skipped = 0;
    if (0) fprintf(stderr, "written\n");
    break;

  case DISCARD_IF_BLOCK:
    network_client_output_fifo_add(nc, buffer, len, 1);
    nbytes = len;		/* say we sent the bytes even though we queued or discarded them them */
    break;

  case QUEUE_IF_BLOCK:
    network_client_output_fifo_add(nc, buffer, len, MAX_BYTES_TO_QUEUE);
    nbytes = len;		/* say we sent the bytes even though we queued them */
    break;


  default:
    fprintf(stderr, "network_client_write: network_blocking_behavior %d unknown", ifBlock);
    exit(-1);
  }

  if (nbytes < 0) {
    fprintf(stderr, "fd=%d len=%d got=%d\n", fd, len, nbytes);
    perror("network_client_write");
    fprintf (stderr,
	     "network_client_write: lost client from host %s, port %hd.\n",
	     inet_ntoa (nc->client_sockaddr_in->sin_addr),
	     ntohs (nc->client_sockaddr_in->sin_port));
    nc->state = CLOSING;
  }
  return nbytes;
}

// Attempt to write out the fifo and abort on first block.
// Keep nc->fifo up to date.
// Return -1 on error with errno set.  EAGAIN means try again.
// Advance to the next buffer only after one buffer gets fully written.
// Start at start and if not fully done, update start and len in the head of the fifo.
// On success, return number of bytes written in most recent queue.
static int network_client_flush_fifo(network_client *nc) {
  int nbytes = 0;
  network_client_output_fifo *nf = nc->fifo;
  if (nf == NULL) 
    return 0;
  if (0) {
    fprintf(stderr, "Flushing fifo for client %d...", nc->fd);
    fflush(stderr);
  }

  int fd = nc->fd;
  if (fcntl(fd, F_SETFL, O_NONBLOCK) < 0) {
    perror("fcntl nonblock");
    return -1;		/* will error out and close connection */
  }

  while (nf != NULL) {
    if (0) fprintf(stderr, "writing %d@%d %c...", nf->len, nf->start, nf->buffer[nf->start]);
    fflush(stderr);
    nbytes = send(nc->fd, nf->buffer + nf->start, nf->len, MSG_DONTWAIT);
    if (nbytes == -1) {
      if (errno == EAGAIN) {
	if (0) fprintf(stderr, "%d bytes left queued\n", nf->len);
	if (nc->network_writes_skipped++ > MAX_NETWORK_WRITES_TO_SKIP) {
	  fprintf(stderr, "network_writes_skipped for %d exceeds %d\n", nc->fd, MAX_NETWORK_WRITES_TO_SKIP);
	  // if we get EAGAIN too many times, assume the client is gone
	  // but we for some reason aren't hearing about it, and close it on our side.
	  return -1;
	}
	nbytes = 0;
	break;	
      } else {
	fprintf(stderr, "%d bytes failed %d \n", nf->len, errno);
	return -1;		/* want to close connection */
	break;
      }
    } else {
      nf->start += nbytes;
      nf->len -= nbytes;
      nc->network_writes_skipped = 0;
      if (0) fprintf(stderr, "%d bytes written, %d to go\n", nbytes, nf->len);
    } 

    if (nf->len == 0) {
      // go for the next one.
      network_client_output_fifo *old = nf;
      nf = nc->fifo=nf->next;
      network_client_output_fifo_free(old);
      if (0) fprintf(stderr, "done\n");
    } else {
      // didn't write the whole thing -- give up
      if (0) fprintf(stderr, "didn't write the whole thing -- give up\n");
      break;
    }
  }
  return nbytes;
}

static void network_client_output_fifo_add (network_client *nc, char *buffer, int len, int maxBytesQueued) {
  if (0) fprintf(stderr, "Queuing %d bytes fd %d\n", len, nc->fd);
  int total = 0;
  network_client_output_fifo *nf = nc->fifo;
  if (nf != NULL) {
    while(nf->next != NULL && ((maxBytesQueued == 0) || total < maxBytesQueued)) {
      total += nf->len;
      nf = nf->next;
    }
  }
  if (0) fprintf(stderr, "Queue total is %d\n", total);

  if (maxBytesQueued > 0 && total > maxBytesQueued) {
    if (1) fprintf(stderr, "Dropping %d bytes fd %d because total=%d and maxBytesQueued=%d\n", len, nc->fd, total, maxBytesQueued);
    return;
  }

  if (nf == NULL) {
    char *newbuf = g_malloc(len);
    if (newbuf == NULL) { perror("network_client_output_fifo_add"); exit(-1); }
    memcpy(newbuf, buffer, len);
    network_client_output_fifo *newfifo = network_client_output_fifo_new(newbuf, len);
    nc->fifo = newfifo;
    if (0) fprintf(stderr, "New fifo for client %d\n", nc->fd);
  } else {
    if (0) fprintf(stderr, "Adding at end\n");
    int oldlen = nf->len;
    int combinedLen = (len + oldlen);
    if (combinedLen < WRITE_COMBINING_WINDOW) {
      if (0) fprintf(stderr, "Write combining %d and %d\n", len, oldlen);
      char *newbuf = g_malloc(combinedLen);
      char *oldbuf = nf->buffer;
      if (newbuf == NULL) { perror("network_client_output_fifo_add"); exit(-1); }
      memcpy(newbuf, oldbuf+nf->start, oldlen);
      memcpy(newbuf+oldlen, buffer, len);
      nf->start = 0;
      nf->len = combinedLen;
      nf->buffer = newbuf;
      g_free(oldbuf);
    } else {
      char *newbuf = g_malloc(len);
      if (newbuf == NULL) { perror("network_client_output_fifo_add"); exit(-1); }
      memcpy(newbuf, buffer, len);
      network_client_output_fifo *newfifo = network_client_output_fifo_new(newbuf, len);
      nf->next = newfifo;
    }
  }
}

int network_read_from_client (network_client *nc, char *buffer, int len) {
  if (nc->state == CLOSING) 
    return 0;
  int fd = nc->fd;
  int nbytes = recv (fd, buffer, len, MSG_WAITALL);
  if (nbytes < 0) {
    /* Read error. */
    perror ("network_read_from_client");
    fprintf (stderr,
	     "network_read_from_client: lost client from host %s, port %hd.\n",
	     inet_ntoa (nc->client_sockaddr_in->sin_addr),
	     ntohs (nc->client_sockaddr_in->sin_port));
    nc->state = CLOSING;
    return nbytes;
  } else if (nbytes == 0) {
    /* End-of-file. */
    fprintf (stderr,
	     "network_read_from_client: EOF %s, port %hd.\n",
	     inet_ntoa (nc->client_sockaddr_in->sin_addr),
	     ntohs (nc->client_sockaddr_in->sin_port));
    nc->state = CLOSING;
    return -1;
  } else {
    /* Data read. */
    return nbytes;
  }
}


int make_socket (unsigned short int port)
{
  int sock;
  struct sockaddr_in name;

  /* Create the socket. */
  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    perror ("socket");
    exit (EXIT_FAILURE);
  }
  // try to fix it so we can restart quickly during debuggin
  {
    struct linger x;
    x.l_onoff = 0;
    int rr = setsockopt (sock, SOL_SOCKET, SO_LINGER, &x, sizeof(x));
    if (rr == -1) {
      perror ("network SO_LINGER failed");
      exit(EXIT_FAILURE);
    }
  }
  // taken from netcat
  {
    int x = 1;
    int rr = setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &x, sizeof (x));
    if (rr == -1) {
      perror ("network reuseaddr failed");
      exit(EXIT_FAILURE);
    }
  }
#ifdef SO_REUSEPORT
  {
    int rr = setsockopt (sock, SOL_SOCKET, SO_REUSEPORT, &x, sizeof (x));
    if (rr == -1) {
      perror ("network reuseport failed");
      exit(EXIT_FAILURE);
    }
  }
#endif


  /* Give the socket a name. */
  name.sin_family = AF_INET;
  name.sin_port = htons (port);
  name.sin_addr.s_addr = htonl (INADDR_ANY);
  int times;
  int code = 0;
  for (times = 0; times < 4; times++) {
    errno = 0;
    code = bind (sock, (struct sockaddr *) &name, sizeof (name));
    fprintf(stderr, "try %d got %d\n", times, code);
    if (code == 0) {
      fprintf(stderr, "success!\n");
      break;
    }
    if (errno != EADDRINUSE) {
      fprintf(stderr, "binding port %d\n", port);
      perror ("bind");
      exit (EXIT_FAILURE);
    }
    sleep(2);
  }
  if (code != 0) {
    fprintf(stderr, "binding port %d\n", port);
    perror ("bind");
    exit (EXIT_FAILURE);
  }

  return sock;
}

static network_client *network_client_new(void) {
  network_client *nc = (network_client *)g_malloc(sizeof(network_client));
  if (nc == NULL) {
    perror("malloc");
    exit(-1);
  }
  struct sockaddr_in *s = (struct sockaddr_in *)g_malloc(sizeof(struct sockaddr_in));
  if (s == NULL) {
    perror("malloc");
    exit(-1);
  }
  nc->next = NULL;
  nc->tx_authorized = 0;
  nc->tx = 0;
  nc->fifo = NULL;
  nc->client_sockaddr_in = s;
  nc->state = NEW;
  nc->fd = -1;
  nc->waterfall_skip_amount = 0;
  nc->waterfall_skip_index = 0;
  nc->waterfall_last_frequency = 0;
  // For some reason we are not getting the initial request from the hiptop.
  nc->waterfall_width = 0;
  nc->waterfall_last_bandwidth = 0;

  nc->miniscope_network_enable = 1;
  nc->miniscope_skip_amount = 11;
  nc->miniscope_skip_index = 0;

  nc->network_writes_skipped = 0;

  nc->mutex = g_mutex_new();

  nc->last_mode = NULL;

  network_clients_n++;
  return nc;
}

static network_client_output_fifo *network_client_output_fifo_new(char *newbuf, int len) {
  network_client_output_fifo *fifo = g_malloc(sizeof(network_client_output_fifo));
  if (fifo == NULL) {
    perror("malloc");
    exit(-1);
  }
  fifo->buffer = newbuf;
  fifo->len = len;
  fifo->start = 0;
  fifo->next = NULL;
  return fifo;
}


void network_update_waterfall_frequency(int x2) {
  if (network_clients != NULL) {
    if (0) fprintf(stderr, "Network update waterfall frequency %d\n", x2);
    int nbytes = 0;
    char cmd = 'F';
    char buffer[8];
    network_client *nc;
    buffer[nbytes++]=cmd;
    buffer[nbytes++]=(x2 / 256);
    buffer[nbytes++]=(x2 & 255);
    for (nc = network_clients; nc != NULL; nc = nc->next) {
      if (nc->waterfall_last_frequency != x2) {
	nc->waterfall_last_frequency = x2;
	network_client_write(nc, buffer, nbytes, QUEUE_IF_BLOCK);
      }
    }
  }
}

void network_update_waterfall_bandwidth(double bw) {
  if (network_clients != NULL) {
    int x2 = ((int)((bw * 1000) + 0.5));
    if (0) fprintf(stderr, "Network update bandwidth %g %d\n", bw, x2);
    int nbytes = 0;
    char cmd = 'B';
    char buffer[4];
    network_client *nc;
    buffer[nbytes++]=cmd;
    buffer[nbytes++]=(x2 / 65536);
    buffer[nbytes++]=(x2 / 256);
    buffer[nbytes++]=(x2 & 255);
    for (nc = network_clients; nc != NULL; nc = nc->next) {
      if (nc->waterfall_last_bandwidth != x2) {
	nc->waterfall_last_bandwidth = x2;
	network_client_write(nc, buffer, nbytes, QUEUE_IF_BLOCK);
      }
    }
  }
}


// Tell clietn about received chars.
void network_send_rx_char(gchar c) {
  guchar buffer[3];
  buffer[0]='C';
  buffer[1]=1;
  buffer[2]=(c);
  network_client *nc;
  for (nc = network_clients; nc != NULL; nc = nc->next) {
    network_client_write(nc, buffer, sizeof(buffer), QUEUE_IF_BLOCK);
  }
}

// echo back to show it was sent.
void network_send_tx_char(gchar c) {
  guchar buffer[3];
  buffer[0]='T';
  buffer[1]=1;
  buffer[2]=(c);
  network_client *nc;
  for (nc = network_clients; nc != NULL; nc = nc->next) {
    network_client_write(nc, buffer, sizeof(buffer), QUEUE_IF_BLOCK);
  }
}


static void network_client_close (network_client *nc) {
  if (nc == NULL) {
    fprintf(stderr, "Cannot close null client\n");
    return;
  }
  if (nc->tx) {
    network_abort();
    fprintf(stderr, "A TX client is closing; pushing the abort button.");
  }
  int fd = nc->fd;
  FD_CLR (fd, &network_active_fd_set);
  close (fd);
  network_client_free(nc);
}

void NULLMUTEX() {
  fprintf(stderr, "null mutex\n");
}

int network_clients_n = 0;
static void network_client_free(network_client *nc) {
  gboolean found = FALSE;
  {
    if (nc == network_clients) {
      network_clients = nc->next;
      found = TRUE;
    } else {
      network_client *p = network_clients; /* we've already checked the first one */
      while (p != NULL) {
	if (p->next == nc) {
	  p->next = nc->next;
	  found = TRUE;
	  break;
	}
	p = p->next;
      }
    }
    if (!found) {
      fprintf(stderr, "network_client_free called on unknown network client\n");
      exit(-1);
    }
    fprintf(stderr, "network: closing client %s\n", inet_ntoa (nc->client_sockaddr_in->sin_addr));

    {
      network_client_output_fifo *fifo = nc->fifo;
      while (fifo != NULL) {
	network_client_output_fifo *next = fifo->next;
	network_client_output_fifo_free(fifo);
	fifo = next;
      }
    }
  }
  if (nc->mutex == NULL) {
    NULLMUTEX();
  }
  g_mutex_unlock(nc->mutex);
  g_mutex_free(nc->mutex);
  nc->mutex = NULL;
  g_free(nc);
  network_clients_n--;
  fprintf(stderr, "network: now %d clients\n", network_clients_n);
}

// just frees it and buffer -- assumes caller has spliced it out.
static void network_client_output_fifo_free(network_client_output_fifo *nf) {
  g_free(nf->buffer);
  g_free(nf);
}


static network_client *network_client_from_fd(int fd) {
  network_client *nc;
  for (nc = network_clients; nc != NULL; nc = nc->next) {
    if (nc->fd == fd) return nc;
  }
  return NULL;
}
